//****************************************************************************
//
// Process the "ASL" instruction 
//
// ASO    ***    (SLO)
// This opcode ASLs the contents of a memory location and then ORs the result 
// with the accumulator.  
// 
// Supported modes:
// 
// ASO abcd        ;0F cd ab    ;No. Cycles= 6
// ASO abcd,X      ;1F cd ab    ;            7
// ASO abcd,Y      ;1B cd ab    ;            7
// ASO ab          ;07 ab       ;            5
// ASO ab,X        ;17 ab       ;            6
// ASO (ab,X)      ;03 ab       ;            8
// ASO (ab),Y      ;13 ab       ;            8
//
//****************************************************************************
void ProcessASO_C( void )
{
	word  address;
	word  zpage;

	_F&=(_V+_D+_I);
	switch (opcode) {
	   case 0x0F	:	address=GETAWORD(); data=Peek(address);     Cycles=6; break;
	   case 0x1F	:	address=GETAWORD()+_X; data=Peek(address);  Cycles=7; break;
	   case 0x1B	:	address=GETAWORD()+_Y; data=Peek(address);  Cycles=7; break;
	   case 0x07	:	address=GETABYTE(); data=Peek(address); 	  Cycles=5; break;
	   case 0x17	:	address=GETABYTE()+_X; data=Peek(address);  Cycles=6; break;
	   case 0x03	:	address=(GETABYTE()+_X)&255; data=Peek(address); Cycles=8; break;
	   case 0x13	:	zpage=GETABYTE(); address=Deek(zpage)+_Y;    data=Peek(address); Cycles=8; break;
	}

	if ((data&128)==128) { _F|=_C; }
	data=data<<1;
	if (data==0)  { _F|=_Z; }
	if ((data&128)==128) { _F|=_N; }

	Poke(address,data);
	_A=(_A|data)&255;
}

