//****************************************************************************
//
// Process the "LSE" instruction    _N=0     0->_A->_C
//
//	LSE    ***   (SRE)
//	LSE LSRs the contents of a memory location and then EORs the result with 
//	the accumulator.
//
//	Supported modes:
//
//	LSE ab          ;47 ab       ;            5
//	LSE ab,X        ;57 ab       ;            6
//	LSE abcd        ;4F cd ab    ;No. Cycles= 6
//	LSE abcd,X      ;5F cd ab    ;            7
//	LSE abcd,Y      ;5B cd ab    ;            7
//	LSE (ab,X)      ;43 ab       ;            8
//	LSE (ab),Y      ;53 ab       ;            8
//
//****************************************************************************
void ProcessLSE_C( void )
{
	word  address;
	byte  zpage;


	_F&=(_V+_D+_I);
	switch (opcode) {
	   case 0x47	:	zpage=GETABYTE(); data=Peek(zpage); 	  break;
	   case 0x57	:	zpage=GETABYTE(); data=Peek((zpage+_X)&255); break;
	   case 0x4F	:	address=GETAWORD(); data=Peek(address); break;
	   case 0x5F	:	address=GETAWORD(); data=Peek(address+_X); break;
	   case 0x5B	:	address=GETAWORD(); data=Peek(address+_Y); break;
	   case 0x43	:	zpage=GETABYTE(); address=Deek((zpage+_X)&255); data=Peek(address);    Cycles=6; break;
	   case 0x53	:	zpage=GETABYTE(); address=Deek(zpage);    data=Peek(address+_Y); Cycles=5; break;
	}

	if ((data&1)==1) { _F|=_C; }
	data=data>>1;
	if (data==0)  { _F|=_Z; }

	switch (opcode) {
	   case 0x47	:	Poke(zpage,data);    Cycles=5; break;
	   case 0x57	:	Poke((zpage+_X)&255,data); Cycles=6; break;
	   case 0x4F	:	Poke(address,data);  Cycles=6; break;
	   case 0x5F	:	Poke(address+_X,data); Cycles=7; break;
	   case 0x5B	:	Poke(address+_Y,data); Cycles=7; break;
	   case 0x43	:	Poke(address,data);    Cycles=8; break;
	   case 0x53	:	Poke(address+_Y,data); Cycles=8; break;
	}
	_A^=data;
}

