//****************************************************************************
//
// Process the "ROL" instruction 
//
//	RLA    ***
//	RLA ROLs the contents of a memory location and then ANDs the result with 
//	the accumulator.
//
//	Supported modes:
//
//	RLA ab          ;27 ab       ;            5
//	RLA ab,X        ;37 ab       ;            6
//	RLA abcd        ;2F cd ab    ;No. Cycles= 6
//	RLA abcd,X      ;3F cd ab    ;            7
//	RLA abcd,Y      ;3B cd ab    ;            7
//	RLA (ab,X)      ;23 ab       ;            8
//	RLA (ab),Y      ;33 ab       ;            8
//
//****************************************************************************
void ProcessRLA_C( void )
{
	word  address;
	byte  zpage;
	byte  t;

	t=0;
	if ((_F&_C)!=0) t=1;
	_F&=(_V+_D+_I);
	switch (opcode) {
	   case 0x27	:	zpage=GETABYTE(); data=Peek(zpage);  break;
	   case 0x37	:	zpage=GETABYTE(); data=Peek((zpage+_X)&255); break;
	   case 0x2F	:	address=GETAWORD(); data=Peek(address); Poke(address,data);break;
	   case 0x3F	:	address=GETAWORD(); data=Peek(address+_X); break;
	   case 0x3B	:	address=GETAWORD(); data=Peek(address+_Y); break;
	   case 0x23	:	zpage=GETABYTE(); address=Deek((zpage+_X)&255); data=Peek(address); break;
	   case 0x33	:	zpage=GETABYTE();	address=Deek(zpage); data=Peek(address+_Y); break;
	}

	
	if ((data&128)==128) { _F|=_C; }
	data=(data<<1)|t;
	if (data==0)  { _F|=_Z; }
	if ((data&128)==128) { _F|=_N; }
	_A=_A&data;

	switch (opcode) {
	   case 0x26	:	Poke(zpage,data);  	Cycles=5; break;
	   case 0x36	:	Poke((zpage+_X)&255,data); Cycles=6; break;
	   case 0x2F	:	Poke(address,data); 	Cycles=6; break;
	   case 0x3F	:	Poke(address+_X,data); Cycles=7; break;
	   case 0x3B	:	Poke(address+_Y,data); Cycles=7; break;
	   case 0x23	:	Poke(address,data); Cycles=8; break;
	   case 0x33	:	Poke(address+_Y,data); Cycles=8; break;
	}
}


