//****************************************************************************
//
// Process the "RRA" instruction 
//
//	RRA    ***
//	RRA RORs the contents of a memory location and then ADCs the result with 
//	the accumulator.
//
//	Supported modes:
//
//	RRA ab          ;67 ab       ;            5
//	RRA ab,X        ;77 ab       ;            6
//	RRA abcd        ;6F cd ab    ;No. Cycles= 6
//	RRA abcd,X      ;7F cd ab    ;            7
//	RRA abcd,Y      ;7B cd ab    ;            7
//	RRA (ab,X)      ;63 ab       ;            8
//	RRA (ab),Y      ;73 ab       ;            8
//
//****************************************************************************
void ProcessRRA_C( void )
{
	word  address;
	byte  zpage;
	byte  t;
	byte	last;
	uint16	d2;

	t=0;
	if ((_F&_C)!=0) t=128;
	_F&=(_V+_D+_I);
	switch (opcode) {
	   case 0x67	:	zpage=GETABYTE(); data=Peek(zpage); 	break;
	   case 0x77	:	zpage=GETABYTE(); data=Peek((zpage+_X)&255); break;
	   case 0x6F	:	address=GETAWORD(); data=Peek(address); break;
	   case 0x7F	:	address=GETAWORD(); data=Peek(address+_X); break;
	   case 0x7B	:	address=GETAWORD(); data=Peek(address+_Y); break;
	   case 0x63	:	zpage=GETABYTE(); address=Deek((zpage+_X)&255); data=Peek(address);  break;
	   case 0x73	:	zpage=GETABYTE(); address=Deek(zpage);    data=Peek(address+_Y); break;
	}

	
	if ((data&1)==1) { _F|=_C; }
	data=(data>>1)|t;
	if (data==0)  { _F|=_Z; }
	if ((data&128)==128) { _F|=_N; }

	switch (opcode) {
	   case 0x67	:	Poke(zpage,data); Cycles=5; break;
	   case 0x77	:	Poke((zpage+_X)&255,data); Cycles=6; break;
	   case 0x6F	:	Poke(address,data); 	Cycles=6; break;
	   case 0x7F	:	Poke(address+_X,data); Cycles=7; break;
	   case 0x7B	:	Poke(address+_Y,data); Cycles=7; break;
	   case 0x63	:	Poke(address,data);  Cycles=8; break;
	   case 0x73	:	Poke(address+_Y,data); Cycles=8; break;
	}


	//
	// Do ADC part
	//
	_d1=_A;
	if ((_F&_C)!=0) _d1++;
	_d1+=data;								// add data with _A

	// do flags
	_F&=(_I+_D);								// clear carry as well
	if ((_d1&0x100)!=0) _F|=_C;
	_A=_d1&255;
	if ((_A & 128)==128) {_F|=_N;}
	if (_A==0) {_F|=_Z;}

}
