//****************************************************************************
//
// Process the "CMP" instruction 
//
//	SAX    ***
//	SAX ANDs the contents of the A and X registers (leaving the contents of A 
//	intact), subtracts an immediate value, and then stores the result in X.
//	... A few points might be made about the action of subtracting an immediate 
//	value.  It actually works just like the CMP instruction, except that CMP 
//	does not store the result of the subtraction it performs in any register.  
//	This subtract operation is not affected by the state of the Carry flag, 
//	though it does affect the Carry flag.  It does not affect the Overflow 
//	flag.
//
//	One supported mode:
//
//	SAX #ab         ;CB ab       ;No. Cycles= 2
//
//	Example:
//
//	SAX #$5A        ;CB 5A
//
//	Equivalent instructions:
//
//	STA $02
//	TXA
//	AND $02
//	SEC
//	SBC #$5A
//	TAX
//	LDA $02
//
//****************************************************************************
void ProcessSAX_C( void )
{
	word  address;
	byte  zpage;
	byte  test,a;


	a=_A&_X;

	data=GETABYTE(); 
	Cycles=2; 

	_F&=(_I+_D+_V+_B);					// wipe flags
	if (a>=data) {_F|=_C;}				// carry set when a>=data (clear = a<data)
	test=a-data;
	_X=test;
	if ((test&128)!=0) {_F|=_N;}		// sign bit set?
	if (test==0) {_F|=_Z;}
}


