//****************************************************************************
//
//  Standard header file.
//
//****************************************************************************
#include <stdio.h>
#include <sys\stat.h>
#include <sys\types.h>
#include <fcntl.h>
#include "malloc.h"


#define	xres	368
#define	yres	240
#define	Xsize	xres
#define	Ysize	yres

//
// flag for memory read and write
// Means that its NOT a memory location,
// but a routine to call!
//
#define	ROUTINE	0x00000001				// all JUMP routines are 4 byte aligned!

#define	word			unsigned short
#define	byte			unsigned char
#define	MAXROM		9

#define  MODEX			1

#define	SCREEN_LINES		0x137						// 312 scan lines

//
// PAL timings. These are now EXACT. Cycle stealing WILL change THESE numbers.
//
// Single cycle mode (895KHz)   57
// Double cycle mode (1.788KHz) 114			
//
#define	CYCLES_PER_LINE_ON_def	57				// ++ NORMAL ON  SCREEN CYCLES ++ 
#define	CYCLES_PER_LINE_OFF_def	114			// ++ NORMAL OFF SCREEN CYCLES ++ 

typedef	struct	tScreenInfo{
			uint32	blank;				// screen blanked???
			uint32	bmm;					// bitmap mode?
			uint32	multicol;			// hardware multicolour mode
			uint32	Invert;
			uint32	CharSet;
			uint32	BitMap;				// Bitmap address
			uint32	ScrollX;
			uint32	ScrollY;
			uint32	ScreenAddress;
			uint32	Cursor;
			uint8		col0;
			uint8		col1;
			uint8		col2;
			uint8		col3;
			uint8		col4;
}tScreenInfo;


// 65816
//#define _V	   1
//#define _D	   2
//#define _I	   4
//#define _C	   8
//#define _Z	   16
//#define _N	   32
//#define _B	   64

// 8501
#define _C	   1
#define _Z	   2
#define _I	   4
#define _D	   8
#define _B	   16
#define _V	   64
#define _N	   128

#define	JOYPORT			0x201			// joystick port
#define	BUTTON_1_A		0x10			// joystick 1 button A
#define	BUTTON_1_B		0x20			// joystick 1 button B
#define	BUTTON_2_A		0x40			// joystick 2 button A
#define	BUTTON_2_B		0x80			// joystick 2 button B

#define	JOYSTICK_1_X	0x01			// joystick 1 X axis
#define	JOYSTICK_1_Y	0x02			// joystick 1 Y axis
#define	JOYSTICK_2_X	0x04			// joystick 2 X axis
#define	JOYSTICK_2_Y	0x08			// joystick 2 Y axis

#define	JOY_1_CAL		1				// cal joystick 1 or 2?
#define	JOY_2_CAL		2


#define	ColourNo		0x3c8
#define	ColourAttr	0x3c9
#define	AND			&&
#define	OR				||
#define	NOT			!=
#define	XOR				^
#define	EOR				^
#define	TRUE			255
#define	FALSE			0

//    1-17   21 sectors * 256 bytes
//   18-24   19 sectors * 256 bytes
//   25-30   18 sectors * 256 bytes
//   31-35   17 sectors * 256 bytes
//  *36-40*  17 sectors * 256 bytes
//					21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,
//					19,19,19,19,19,19,19,
//					18,18,18,18,18,18,
//					17,17,17,17,17,
//					17,17,17,17,17,



typedef	struct tDir {
		uint8		FileType;
		uint8		Track;
		uint8		Sector;
		char		FileName[16];		// NO 0x0 end
		uint8		RelTrack;
		uint8		RelSector;
		uint8		RelRecordLen;
		uint8		Unused[4];
		uint8		OverwriteTrack;
		uint8		OverwriteSector;
		uint16	NumberBlocks;
		uint8		unused2;				// skip bytes
		uint8		unused3;
}tDir;

typedef	struct	tSector{
		uint8		Track;				// Next track
		uint8		Sector;				// and sector
		uint8		Data[254];			// 254 bytes of data.
}tSector;

typedef	struct	tMenu{
		 uint8	FileType;			// selectable or not (0 or 1)
		 char		Text[41];			// message to display
		 uint8	Track; 				// track and
		 uint8	Sector;				// sector of program
struct tMenu	*pNext;				// next menu item in chain
}tMenu;

