//****************************************************************************
//
//
//
//        The Minus4 project - Copyright (c) Mike Dailly 1996,7
//
//
// This source and the Minus4 project remains the copyright of Mike Dailly, 
// but can be freely used and changed for NON-Commercial purposes. 
//
//								 Keep emulators free!
//
// Remember if you add or change anything, you MUST send a copy
// complete with source to me at mike@scotch.demon.co.uk
// If its just a bug fix, then just send the changed file, with a comment
// telling me WHERE you fixed and WHAT you fixed.
//
// 
//										 SOUND.C
//                                      
//	This file holds the ADLIB (FM) SOUND stuff. It is used to emulate the 
// Plus/4's "internal" sound produced by the TED chip.
//		
//
// Version  By   Date		Description
//  V0.01	MD	   ?			File started. Not much more..
//  V0.02   MD  9/10/97		AdLib sound added for internal Plus/4 sound/Ted chip
//                                 		 
//
//
//****************************************************************************
#include "headers\md.h"
#include "headers\all.h"
#include "headers\screen.h"
#include "headers\draw.h"
#include "headers\6502.h"
#include "headers\6502asm.h"
#include "headers\disasm.h"
#include "headers\keyb.h"
#include	"headers\keyboard.h"

#define	AddressPort	0x388					// also the status register
#define	DataPort		0x389
#define	RegisterDelay	6
#define	DataDelay		35
#define	AdLibConvert	1.3113636f		// adlib frequ to register value convertor

	uint32	AdLibPort=AddressPort;
	uint32	SOUND=1;							// sound active? (yes=default)

	uint8		_ff0e=0,_ff0f=0,_ff10=0,_ff11=0,_ff12=0;		// check for changes
	uint16	Chan1,Chan1Oct,Chan2,Chan2Oct;
	uint32	AdLibReg,Octive;
	uint32	Lastff11;
	uint32	Chan1ON=0,Chan2ON=0;

	uint32	VolumeCounter;
	uint8		a,b;
	uint16	ab;
	float		freq;
	uint8		flag;

	uint16	v;

//************************************************************
//                                                           *
//   Name    : AdLibDelay                                    *
//   Function: After writting to the ADLIB port, we need a   *
//             delay.. this is it! 6 for a register select   *
//             and 35 for a data write.                      *
//                                                           *
//   In   : delay = number of in's to do                     *
//   Out  : none                                             *
//                                                           *
//************************************************************

void AdLibDelay(uint32 val  );
#pragma aux AdLibDelay =				\
		"mov	edx,AdLibPort"				\
"MyLp: in	al,dx"						\
		"dec	ecx"							\
		"jne	MyLp"							\
	modify [al dx ecx]					\
	parm	[ecx]


//************************************************************
//                                                           *
//   Name    : GetFreq													 *
//   Function: Take a Plus/4 register and return the real    *
//             frequency its playing                         *
//                                                           *
//   In   : SoundReg = Plus/4 0-$3ff sound value             *
//   Out  : float    = actual frequency.                     *
//                                                           *
//************************************************************

float	GetFreq ( uint16 SoundReg )
{
	float	reg;

	reg=(float) (1024-SoundReg);
	reg=111860.781/reg;							// NTSC timings.
	return reg;										// return "HZ" hertz
}

//************************************************************
//                                                           *
//   Name    : GetAdLibReg                                   *
//   Function: enter with REAL frequence and return with     *
//             Adlib octive and register setting             *
//                                                           *
//   In   : freq = frequence to convert                      *
//   Out  : AdLibReg= value to use                           *
//          Octive  = octive to play at                      *
//                                                           *
//************************************************************

void GetAdLibReg ( float freq )
{
	int	Oct=4;
   int 	block,b,m,f,fn;

	if (freq>780){
		while (freq>780){
			freq=freq/2;
			Oct++;
		}	
	}

	freq=freq*1.3113636;
	AdLibReg=(uint32) freq;
	Octive=Oct;
}


//************************************************************
//                                                           *
//   Name    : WriteAddress                                  *
//   Function: Write out to the adlib address port and wait  *
//                                                           *
//   In   : Reg = register to select                         *
//   Out  : none                                             *
//                                                           *
//************************************************************

void WriteAddress (uint8 Reg)
{
	outp (AddressPort,Reg);			// select register 4
	AdLibDelay (RegisterDelay);
}					

//************************************************************
//                                                           *
//   Name    : WriteData                                     *
//   Function: write a value to the data port and wait.      *
//                                                           *
//   In   : Val = value to write                             *
//   Out  : void                                             *
//                                                           *
//************************************************************

void WriteData (uint8 Val)
{
	outp (DataPort,Val);			// select register 4
	AdLibDelay (DataDelay);
}		 

//************************************************************
//                                                           *
//   Name    : Reg                                           *
//   Function: Write a value to a register                   *
//                                                           *
//   In   : Reg = register to write to                       *
//          Val = value to write                             *
//   Out  : none                                             *
//                                                           *
//************************************************************

void	Reg (uint8 Port, uint8 Val){
	WriteAddress(Port);
	WriteData(Val);
}

//************************************************************
//                                                           *
//   Name    : ReadStatus                                    *
//   Function: read from adlib status register               *
//                                                           *
//   In   : none                                             *
//   Out  : status                                           *
//                                                           *
//************************************************************

uint8 ReadStatus ( void )
{
	uint8	a;
	a=inp (AddressPort);	
	AdLibDelay (RegisterDelay);		// need this?
	return a;
}					

//************************************************************
//                                                           *
//   Name    : DetectSoundAdLib                              *
//   Function: detect if we have an adlib sound card         *
//                                                           *
//   In   : none                                             *
//   Out  : 1 = yes, 0 = no                                  *
//                                                           *
//************************************************************

int DetectSoundAdLib( void )
{
	uint8	a,b,c;

	WriteAddress(4);
	WriteData   (0x60);
	WriteAddress(4);
	WriteData   (0x80);
	a=ReadStatus();	
	WriteAddress(2);
	WriteData   (0xff);
	WriteAddress(4);
	WriteData   (0x21);
	AdLibDelay(200);			// wait at LEAST 80 microseconds
	b=ReadStatus();
	WriteAddress(4);
	WriteData   (0x60);
	WriteAddress(4);
	WriteData   (0x80);
	a=a&0xe0;
	b=b&0xe0;
	if ((a==0) AND (b==0xc0))	// detected?
		return 1;
	else
		return 0;
}


//************************************************************
//                                                           *
//   Name    : MakeSound                                     *
//   Function: Make the basic waveforms for our sounds       *
//                                                           *
//************************************************************
void	MakeSound( void ){
	Reg(0x20,0x01);		// setup channel 1 waveform
	Reg(0x40,0x10);
	Reg(0x60,0xF0);
	Reg(0x80,0x77);
	Reg(0x23,0x01);
	Reg(0x43,0x00);
	Reg(0x63,0xF0);
	Reg(0x83,0x77);

	Reg(0x21,0x20);		// set up channel 2 waveform
	Reg(0x41,0x10);
	Reg(0x61,0xF0);
	Reg(0x81,0x77);
	Reg(0x24,0x20);
	Reg(0x44,0x00);
	Reg(0x64,0xf0);
	Reg(0x84,0x77);


	Reg(0x34,0x01);		// set up NOISE channel(s)
	Reg(0x54,0x00);		
	Reg(0x74,0xF0);
	Reg(0x94,0xFF);

	Reg(0x35,0x01);		
	Reg(0x55,0x00);		
	Reg(0x75,0xF0);
	Reg(0x95,0x77);

	Reg(0x31,0x01);		
	Reg(0x51,0x00);		
	Reg(0x71,0xF0);
	Reg(0x91,0xFF);

	Reg(0xBD,0x00);		// Disable Rhythm(noise) just now...

}

//************************************************************
//                                                           *
//   Name    : Chanel1                                       *
//   Function: Update chanel 1                               *
//                                                           *
//   In   : none. actually it reads the plus/4's chanel 1    *
//          register.                                        *
//   Out  : none                                             *
//                                                           *
//************************************************************

void Chanel1( void )
{	
	if (SOUND==0) return;

	a=Memory[0xff0e];							// get sound channel 1
	b=Memory[0xff12]&3;						// get sound channel 1
	if ((a!=_ff0e) OR (b!=_ff12)) {		// channel changed?
		_ff0e=a;
		_ff12=b;
		ab=(b<<8)+a;							// make sound register value.
		freq=GetFreq(ab);						// convert to frequency.
		GetAdLibReg(freq);					// work out freq and oactive
		Chan1=(uint16) AdLibReg;			// remember actual values
		Chan1Oct=(uint16) Octive; 			// base it round octive 3 and up
	}

	if (_ff11==Memory[0xff11]){						// same as LAST time?
		if ((Memory[0xff11]&0x10)==0) return;		// if it WAS and the channel was OFF
		if ((Memory[0xff11]&0x0f)==0) return;		// if it WAS and the channel was OFF
	}

	if ((_ff11&0x80)!=0) return;

	if ((Memory[0xff11]&0x10)!=0){	// channel on?
		v=Chan1+(Chan1Oct<<10);
		Reg(0xA0,v&0xff);	  				// set freq and play noise!
		Reg(0xB0,0x20+((v>>8)&0xff));
		Chan1ON=1;
	} else {
		Reg(0xB0,0);		// channel off
		Chan1ON=0;
	}
}

//************************************************************
//                                                           *
//   Name    : Chanel2                                       *
//   Function: Update chanel 2                               *
//                                                           *
//   In   : none. as above, but chanel 2!							 *
//   Out  : none                                             *
//                                                           *
//************************************************************
void Chanel2( void )
{	
	if (SOUND==0) return;

	a=Memory[0xff0f];							// get sound channel 1
	b=Memory[0xff10]&3;						// get sound channel 1
	if ((a!=_ff0f) OR (b!=_ff10)) {		// channel changed?
		_ff0f=a;
		_ff10=b;
		ab=(b<<8)+a;							// make sound register value.
		freq=GetFreq(ab);						// convert to frequency.
		GetAdLibReg(freq);					// work out freq and oactive
		Chan2=(uint16) AdLibReg;			// remember actual values
		Chan2Oct=(uint16) Octive; 			// base it round octive 3 and up
	}

	if (_ff11==Memory[0xff11]){					// same as LAST time?
		if ((Memory[0xff11]&0x60)==0)	return;	// if it WAS and the channel was OFF
		if ((Memory[0xff11]&0x0f)==0)	return;	// if it WAS and the channel was OFF
	}

	if ((_ff11&0x80)!=0) return;
	
	if ((Memory[0xff11]&0x20)!=0){	// channel on?
		v=Chan2+(Chan2Oct<<10);
		Reg(0xA1,v&0xff);	  				// set freq and play noise!
		Reg(0xB1,0x20+((v>>8)&0xff));
		Reg(0xBD,0x00);					// NOT noise... disable.
		Chan2ON=1;
	} else if ((Memory[0xff11]&0x40)!=0){	// channel on?
		Reg(0xBD,0x28+1+2);					// PLAY noise. No freq control for noise.
		Reg(0xB1,0);							// Normal channel off
		Chan2ON=1;
	} else {
		Reg(0xB1,0);						// channel off
		Reg(0xBD,0x00);					// Rhythm active, and use SNARE.. close to "noise"
		Chan2ON=0;
	}
}

//************************************************************
//                                                           *
//   Name    : ChanelSelect                                  *
//   Function: Chanel select register processing             *
//                                                           *
//   In   : Plus/4 chanel select register                    *
//   Out  :                                                  *
//                                                           *
//************************************************************
void ChanelSelect ( void )
{
	uint8	ff11;

	if (SOUND==0) return;

	ff11=Memory[0xff11];

	// see how many times the "volume" has been clicked this frame.
	// if over 4, then chances are we want to play a sample!
	if ((_ff11&0xf)!=(ff11&0xf)){
		VolumeCounter++;
	}

	// first check for SAMPLE mode
//	if (((ff11&0x80)!=0) OR ((ff11&0xf)==0)) {

	if ((ff11&0x80)!=0) {
		if (Chan1ON==1){
			Reg(0xB0,0);		// channel off
			Chan1ON=0;
		}
		if (Chan2ON==1){
				Reg(0xB1,0);						// channel off
				Reg(0xBD,0x00);					// Rhythm active, and use SNARE.. close to "noise"
				Chan2ON=0;
		}
		_ff11=ff11;
		return;
	}

	if (Memory[0xff0e]|Memory[0xff0f]|Memory[0xff10]|Memory[0xff12]!=0){
		if ((ff11&0x10)!=(_ff11*0x10))	Chanel1();
		if ((ff11&0x60)!=(_ff11*0x60))	Chanel2();
	}

	_ff11=ff11;

}	

//************************************************************
//                                                           *
//   Name    : InitSound                                     *
//   Function: Init all sound stuff                          *
//                                                           *
//   In   : none                                             *
//   Out  : SOUND = 1 (active) or 0 (inactive)               *
//                                                           *
//************************************************************

void InitSound( void )
{
	int	i;

	if (SOUND==1){
		printf("Init Sound System:\n");
		if (DetectSoundAdLib ()==1){
			printf("     Detected ADLIB compatable sound card\n");

			for (i=1;i<0xf6;i++){
				WriteAddress (i);			// clear and init adlib card.
				WriteData 	 (0);
			}
			MakeSound();
		} else {
			printf("     Can't find an ADLIB compatable sound card - Sound disabled\n");
			SOUND=0;
		}
	} else {
			printf("     Audio DISABLED.\n");

	}
}



//************************************************************
//                                                           *
//   Name    : ShutDownSound                                 *
//   Function: Guess what!                                   *
//                                                           *
//************************************************************

void ShutDownSound( void )
{
	int	i;
	
	for (i=1;i<0xf6;i++){
		WriteAddress (i);			// clear and init adlib card.
		WriteData 	 (0);
	}
}


